package sshwrapper

import (
	"bytes"
	"context"
	"fmt"
	"log"
	"os/exec"
	"path/filepath"
	"strings"
	"time"
)

type SSHCommandBackend struct {
	Timeout    time.Duration
	ConnOpts   ConnOpts
	ClientName string
}

func (l *SSHCommandBackend) command(args []string, env map[string]string) []string {
	base := l.ConnOpts.CommandBase()
	e := make([]string, 0)

	for k, v := range env {
		e = append(e, fmt.Sprintf("%s=%s", k, v))
	}

	cmd := append(
		base,
		e...,
	)
	return append(
		cmd,
		args...,
	)
}

func (l *SSHCommandBackend) RunRemoteCommand(parent context.Context, directArgs []string, env map[string]string, stdin string) (string, error) {
	ctx, cancel := context.WithTimeout(parent, l.Timeout)
	defer cancel()

	args := l.command(directArgs, env)
	c := exec.CommandContext(ctx, args[0], args[1:]...)
	e := make([]string, 0)

	for k, v := range env {
		e = append(e, fmt.Sprintf("%s=%s", k, v))
	}

	c.Env = e

	finished := make(chan error)
	output := &bytes.Buffer{}

	c.Stdin = strings.NewReader(stdin)
	c.Stdout = output
	c.Stderr = output

	cmdArgs := strings.Join(args, " ")

	log.Println(&Error{Name: l.ClientName, Message: fmt.Sprint("[", cmdArgs, "]")})
	err := c.Start()
	if err != nil {
		return "", &Error{Name: l.ClientName, Message: fmt.Sprint(cmdArgs, " => could not start process"), Err: err}
	}

	go func() {
		err := c.Wait()
		select {
		case <-ctx.Done():
			return
		default:
			finished <- err
		}
	}()

	select {
	case err := <-finished:
		out := output.String()
		if err != nil {
			return out, &Error{Name: l.ClientName, Message: fmt.Sprint(cmdArgs, " => \"", out, "\""), Err: err}
		}
		return out, nil
	case <-ctx.Done():
		return "", &Error{Name: l.ClientName, Message: fmt.Sprint(cmdArgs, " => process timed out")}
	}
}

func (l *SSHCommandBackend) CmkRunCommand(ctx context.Context, hostUUID string, directArgs []string, env map[string]string) (string, error) {
	if res, err := l.RunRemoteCommand(ctx, directArgs, env, ""); err != nil {
		if e, ok := err.(*Error); ok {
			e.Name = l.ClientName + ":" + hostUUID
		}
		return res, err
	} else {
		return res, err
	}
}

func (l *SSHCommandBackend) CmkWriteFile(ctx context.Context, hostUUID string, path string, content []byte) error {
	remoteDir := filepath.Dir(path)
	log.Println("create remote path:", remoteDir)
	if _, err := l.RunRemoteCommand(ctx, []string{"mkdir", "-p", remoteDir}, map[string]string{}, ""); err != nil {
		if e, ok := err.(*Error); ok {
			e.Name = l.ClientName + ":" + hostUUID
		}
		return err
	}

	if _, err := l.RunRemoteCommand(ctx, []string{"bash", "-c", fmt.Sprint("cat > ", path)}, map[string]string{}, string(content)); err != nil {
		if e, ok := err.(*Error); ok {
			e.Name = l.ClientName + ":" + hostUUID
		}
		return err
	}

	return nil
}

// Add single quotes around argument to execute it via SSH
func (l *SSHCommandBackend) QuoteCommandArg(arg string) string {
	return "'" + strings.Replace(arg, "'", "\\'", -1) + "'"
}
