package transfer

import (
	"context"
	"fmt"
	"log"
	"path/filepath"
	"sync"
	"time"

	"github.com/it-novum/nsta/agent"
	"github.com/it-novum/nsta/checkmk"
	"github.com/it-novum/nsta/gearman"
	"github.com/it-novum/nsta/nstautil"
	"github.com/it-novum/nsta/oitcdb"
	"github.com/it-novum/nsta/satstatus"
	"github.com/it-novum/nsta/sshwrapper"
)

// SSHSatellite manages the connection and message exchange to a satellite via ssh
type SSHSatellite struct {
	// Configuration

	SatelliteOptions

	// Thread Sync

	wg       sync.WaitGroup
	reload   chan chan struct{}
	shutdown chan struct{}

	// SSH + Gearman Connection

	satGearman    *gearman.Client
	ssh           *sshwrapper.Conn
	portPublisher <-chan int

	// State and internal Config

	nagiosConfigPath string
	lastMainExport   time.Time
	toSat            gearman.SyncMessageFIFO

	// Checkmk Commands

	cmk        checkmk.Checkmk
	cmdBackend *sshwrapper.SSHCommandBackend

	agent agent.Agent
}

func (s *SSHSatellite) SatelliteID() int {
	return s.DB.ID
}

func (s *SSHSatellite) SatelliteName() string {
	return s.DB.Name
}

// Compare checks if the satellite matches the database config
func (s *SSHSatellite) Compare(dbSat *oitcdb.Satellite) bool {
	return s.DB.Compare(dbSat)
}

// SendMessage to the satellite
func (s *SSHSatellite) SendMessage(msg *gearman.Message) {
	s.toSat.Push(msg)
}

func (s *SSHSatellite) resetPort(ctx context.Context, port int) {
	status := &satstatus.Status{
		SatelliteID: s.DB.ID,
		Status:      oitcdb.SatelliteStatusFailed,
	}
	defer func() {
		// Update satelite status table
		s.StatusTracker.Update <- status
	}()

	// Reset gearman satellite connection if already connected in the past
	if s.satGearman != nil {
		s.satGearman.Stop()
		s.satGearman = nil
	}

	// if port == 0: Port forwarding or connection was not successfully (or not yet established)
	if port != 0 {
		// Port forwarding successfully, connect to gearman server on the satellite through ssh with a dynamic port
		s.satGearman = &gearman.Client{
			ClientOptions: gearman.ClientOptions{
				ClientType:     gearman.Satellite,
				ClientName:     s.DB.Name,
				Address:        fmt.Sprintf("localhost:%d", port),
				MessageReceive: s.ToMaster,
			},
		}
		s.satGearman.Start(ctx)
		status.Status = oitcdb.SatelliteStatusOK
		status.LastSeen = time.Now()
	}
}

func (s *SSHSatellite) recheckStatus() {
	status := &satstatus.Status{
		SatelliteID: s.DB.ID,
		Status:      oitcdb.SatelliteStatusFailed,
	}
	if s.satGearman != nil && s.satGearman.IsRunning() {
		status.Status = oitcdb.SatelliteStatusOK
		status.LastSeen = time.Now()
	}

	// Update satelite status table
	s.StatusTracker.Update <- status
}

func (s *SSHSatellite) export(parent context.Context) {
	// Copy new monitoring configs to Satellite System and trigger restart-monitoring.sh
	ctx, cancel := context.WithTimeout(parent, time.Second*time.Duration(s.SSHReloadTimeoutSec))
	defer cancel()

	if !s.DB.NSTASyncInstance {
		log.Println(&Error{Name: s.DB.Name, Message: "skip export"})
		return
	}

	s.lastMainExport = time.Now()
	status := &satstatus.Status{
		SatelliteID: s.DB.ID,
	}

	defer func() {
		s.StatusTracker.Update <- status
	}()
	log.Println(&Error{Name: s.DB.Name, Message: "copy configuration"})
	tarData := nstautil.CreateExportTar(s.NagiosPath, s.DB.ID, false)
	// tarData = tar.gz data
	if tarData != "" {
		// Delete old config files
		fullNagiosConfigPathToRemove := filepath.Join(s.nagiosConfigPath, "config")
		if _, err := s.cmdBackend.RunRemoteCommand(ctx, []string{"rm", "-rf", fullNagiosConfigPathToRemove}, map[string]string{}, ""); err != nil {
			status.Err = &Error{Name: s.DB.Name, Message: "removal of old config folder faild", Err: err}
			fmt.Println(status.Err)
			return
		}

		log.Println(&Error{Name: s.DB.Name, Message: "untar new monitoring config"})
		//                                                       tar command reads tarData from stdin
		if _, err := s.cmdBackend.RunRemoteCommand(ctx, []string{"tar", "xzp", "-C", s.nagiosConfigPath}, map[string]string{}, tarData); err != nil {
			status.Err = &Error{Name: s.DB.Name, Message: "export failed", Err: err}
			fmt.Println(status.Err)
			return
		}
		log.Println(&Error{Name: s.DB.Name, Message: "restart monitoring"})
		if _, err := s.cmdBackend.RunRemoteCommand(ctx, []string{"sudo", "/opt/openitc/nagios/bin/restart-monitoring.sh"}, map[string]string{}, ""); err != nil {
			status.Err = &Error{Name: s.DB.Name, Message: "export failed", Err: err}
			fmt.Println(status.Err)
			return
		}
		status.LastExport = time.Now()
		status.LastSeen = time.Now()
		log.Println(&Error{Name: s.DB.Name, Message: "export successful"})
	} else {
		status.Err = &Error{Name: s.DB.Name, Message: "could not create export archive"}
		log.Println(status.Err)
	}
}

// Run is the background go worker
func (s *SSHSatellite) run(ctx context.Context) {
	statusT := time.NewTicker(time.Second * 10) // Update Satellite Status database every 10 seconds
	noNotify := make(chan struct{})
	var timesyncChan <-chan time.Time
	var timesyncTicker *time.Ticker

	fakeTimesync := make(chan time.Time)
	if s.DB.UseTimesync {
		timesyncTicker = time.NewTicker(time.Hour)
		defer timesyncTicker.Stop()
		timesyncChan = timesyncTicker.C
	} else {
		timesyncChan = fakeTimesync
	}

	defer func() {
		log.Println(&Error{Name: s.DB.Name, Message: "shutting down"})
		statusT.Stop()
		if s.satGearman != nil {
			log.Println(&Error{Name: s.DB.Name, Message: "close gearman connection"})
			s.satGearman.Stop()
			s.satGearman = nil
		}
		log.Println(&Error{Name: s.DB.Name, Message: "stop ssh connection"})
		s.ssh.Stop()
		log.Println(&Error{Name: s.DB.Name, Message: "close message fifo"})
		s.toSat.Close()
		log.Println(&Error{Name: s.DB.Name, Message: "shutdown complete"})
		//s.wg.Done()
	}()

	for {
		timesync := timesyncChan
		notify := s.toSat.Notify()
		if s.satGearman == nil {
			timesync = fakeTimesync
			notify = noNotify
		}
		select {
		case <-ctx.Done():
			return
		case _, more := <-s.shutdown:
			if !more {
				return
			}
		case reload := <-s.reload:
			// Got reload signal for this satellite thread
			s.export(ctx)
			reload <- struct{}{} // send responses to main thread that the reload (export) is done
		case <-statusT.C:
			// Update Satellite Status in database
			s.recheckStatus()
		case port := <-s.portPublisher:
			// got new SSH port - restart gearman connection
			s.resetPort(ctx, port)
		case <-notify:
			// Go notify signal from FIFO
			// Transfer message from master to satellite
			msg := s.toSat.Get()
			if msg != nil {
				if msg.Type == gearman.MessageCheckmk {
					//This is a Checkmk message - run SSH command on the satellite
					s.wg.Add(1)
					go func() {
						defer s.wg.Done()
						response, err := s.cmk.ProcessMessage(s.cmdBackend, msg).Serialize()
						if err != nil {
							log.Println(&Error{Name: s.DB.Name, Message: "could no serialize checkmk answer:", Err: err})
						}

						// Timeout for Gearman
						timeout := time.NewTimer(time.Second * 10)
						defer timeout.Stop()

						select {
						case s.ToMaster <- &gearman.Message{
							Type:   gearman.MessageCheckmk,
							String: response,
						}:
						case <-timeout.C:
							log.Println(&Error{Name: s.DB.Name, Message: "could not send checkmk response to gearman: timeout"})
						case <-ctx.Done():
							log.Println(&Error{Name: s.DB.Name, Message: "could not send checkmk response to gearman", Err: ctx.Err()})
						}
					}()
				} else if msg.Type == gearman.MessageOitcAgent {
					//This is a oITC Agent message - execute an agent discovery via SSH on the Satellite system
					s.wg.Add(1)
					go func() {
						defer s.wg.Done()
						response, err := s.agent.ProcessMessage(s.cmdBackend, msg).SerializeOitcAgent()
						if err != nil {
							log.Println(&Error{Name: s.DB.Name, Message: "could no serialize checkmk answer:", Err: err})
						}

						// Timeout for Gearman
						timeout := time.NewTimer(time.Second * 10)
						defer timeout.Stop()

						select {
						case s.ToMaster <- &gearman.Message{
							Type:   gearman.MessageOitcAgent,
							String: response,
						}:
						case <-timeout.C:
							log.Println(&Error{Name: s.DB.Name, Message: "could not send checkmk response to gearman: timeout"})
						case <-ctx.Done():
							log.Println(&Error{Name: s.DB.Name, Message: "could not send checkmk response to gearman", Err: ctx.Err()})
						}
					}()
				} else {
					// Pass external command to Naemon on the Satellite
					s.satGearman.SendMessage(msg)
				}
			}
		case <-timesync:
			// Sync system clock to sattelite system
			s.wg.Add(1)
			go func() {
				defer s.wg.Done()

				tsCtx, tsCancel := context.WithTimeout(ctx, time.Second*60)
				defer tsCancel()

				if _, err := s.cmdBackend.RunRemoteCommand(tsCtx, []string{"sudo", "date", "-s", fmt.Sprint("@", time.Now().Unix())}, map[string]string{}, ""); err != nil {
					log.Println(err)
				} else {
					log.Println(&Error{Name: s.DB.Name, Message: "time update was successful"})
				}
			}()
		}
	}
}

// Close stops the satellite connection
func (s *SSHSatellite) Close() {
	close(s.shutdown)
	s.wg.Wait()
}

// Reload implements a process reload signal handler
func (s *SSHSatellite) Reload(ctx context.Context) {

	// Create a new channel, which gets send to s.reload so s.reload can respond on this channel
	reload := make(chan struct{})

	select {
	case <-ctx.Done():
		return
	case s.reload <- reload: // send responses channel to satellite thread
	}

	// Wait for responses
	select {
	case <-ctx.Done():
	case <-reload:
	}

}

func (s *SSHSatellite) Start(parent context.Context) error {
	s.ssh = &sshwrapper.Conn{
		ConnOpts:   s.SSHConnOpts,
		ClientName: s.DB.Name,
	}

	// Start new SSH Wrapper
	ctx, cancel := context.WithCancel(parent)
	portPub, err := s.ssh.Start(ctx)
	if err != nil {
		cancel()
		return err
	}
	s.portPublisher = portPub // channel used to get the dynamic port from ssh port forwarding

	s.nagiosConfigPath = filepath.Join(s.NagiosPath, "etc")
	s.lastMainExport = time.Now()
	s.cmdBackend = &sshwrapper.SSHCommandBackend{
		Timeout:    time.Second * 120,
		ConnOpts:   s.SSHConnOpts,
		ClientName: s.DB.Name,
	}
	s.cmk.SatelliteName = s.DB.Name
	s.shutdown = make(chan struct{})
	s.reload = make(chan chan struct{})

	s.toSat.Start()
	s.wg.Add(1)
	go func() {
		defer s.wg.Done()
		defer cancel()
		s.run(ctx)
	}()

	return nil
}
